/** \file src/xtal/xtal_cast.h
* \brief src/xtal/xtal_cast.h
*/

#ifndef XTAL_CAST_H_INCLUDE_GUARD
#define XTAL_CAST_H_INCLUDE_GUARD

#pragma once

namespace xtal{

const char_t* cast_const_char_t_ptr(const AnyPtr& a);
IDPtr cast_IDPtr(const AnyPtr& a);

/////////////////////////////////////////////////////////////////////////////

template<class T> struct NullGetter{ static T get(){ return 0; } };
template<class T> struct NullGetter<const SmartPtr<T>&>{ static const SmartPtr<T>& get(){ return nul<T>(); } };
template<class T> struct NullGetter<SmartPtr<T> >{ static const SmartPtr<T>& get(){ return nul<T>(); } };

/////////////////////////////////////////////////////////////////////////////

template<class T>
struct Caster{
	typedef typename Caster<const T&>::type type;
	static type cast(const AnyPtr& a){ return Caster<const T&>::cast(a); }
};

template<class T>
struct Caster<T&>{
	typedef T& type;
	static type cast(const AnyPtr& a){ return (type)Caster<const T&>::cast(a); }
};

template<class T>
struct Caster<const T*>{
	typedef const T* type;
	static type cast(const AnyPtr& a){ return &Caster<const T&>::cast(a); }
};

template<class T>
struct Caster<T*>{
	typedef T* type;
	static type cast(const AnyPtr& a){ return (type)(&Caster<const T&>::cast(a)); }
};

template<class T>
struct Caster<const T&>{
	typedef const T& type;
	static type cast(const AnyPtr& a){ return *reinterpret_cast<const SmartPtr<T>*>(&a)->get(); }
};

template<class T>
struct Caster<const SmartPtr<T>&>{
	typedef const SmartPtr<T>& type;
	static type cast(const AnyPtr& a){ return *reinterpret_cast<const SmartPtr<T>*>(&a); }
};

template<>
struct Caster<const char_t*>{
	typedef const char_t* type;
	static type cast(const AnyPtr& a){ return cast_const_char_t_ptr(a); }
};

template<>
struct Caster<const IDPtr&>{
	typedef IDPtr type;
	static type cast(const AnyPtr& a){ return cast_IDPtr(a); }
};

template<>
struct Caster<const bool&>{
	typedef bool type;
	static type cast(const AnyPtr& a){ return a; }
};

#define XTAL_CAST_HELPER(Type, XType, Conv) \
template<>\
struct Caster<const Type&>{\
	typedef Type type;\
	static type cast(const AnyPtr& a){ return (type)Conv(a); }\
};\
template<>struct Caster<Type&>{\
	typedef Type& type;\
private:\
	static type cast(const AnyPtr& a);\
}

// ȉ̌^͎Qƌ^ɃLXgłȂBconstQƌ^ɂ̓LXgłB

XTAL_CAST_HELPER(char, Int, ivalue);
XTAL_CAST_HELPER(signed char, Int, ivalue);
XTAL_CAST_HELPER(unsigned char, Int, ivalue);
XTAL_CAST_HELPER(short, Int, ivalue);
XTAL_CAST_HELPER(unsigned short, Int, ivalue);
XTAL_CAST_HELPER(int, Int, ivalue);
XTAL_CAST_HELPER(unsigned int, Int, ivalue);
XTAL_CAST_HELPER(long, Int, ivalue);
XTAL_CAST_HELPER(unsigned long, Int, ivalue);
XTAL_CAST_HELPER(float, Float, fvalue);
XTAL_CAST_HELPER(double, Float, fvalue);
XTAL_CAST_HELPER(long double, Float, fvalue);

#undef XTAL_CAST_HELPER


//////////////////////////////////////////////////////////////////////////////

/// \name ^ϊ
//@{

/**
* \brief T^ɕϊł邩ׂB
*
*/
template<class T>
inline bool can_cast(const AnyPtr& a){
	return a->is(&CppClassSymbol<T>::value);
}

/**
* \brief T^Ƀ`FbNŕϊB
*
*/
template<class T>
inline typename Caster<T>::type 
unchecked_cast(const AnyPtr& a){
	return Caster<T>::cast(a);
}

/**
* \brief T^ɕϊB
*
* TɕϊłȂꍇA
* T|C^^Ȃ(T*)NULLԂB
* TlQƂŁAAnyPtrp^Ȃxtal::nullԂB
* ȊǑ^̏ꍇ̓RpCG[ƂȂB
*/
template<class T>
typename Caster<T>::type 
cast(const AnyPtr& a){
	if(can_cast<T>(a)){
		return unchecked_cast<T>(a);
	}
	else{
		return NullGetter<typename Caster<T>::type>::get();
	}
}

/**
* \brief SmartPtr<T>^ɁAۂ̌^ǂł邩𖳎ċϊB
*/
template<class T>
inline const SmartPtr<T>&
unchecked_ptr_cast(const AnyPtr& a){
	return *reinterpret_cast<const SmartPtr<T>*>(&a);
}

/**
* \brief SmartPtr<T>^ɕϊB
*
* TɕϊłȂꍇxtal::nullԂB
*/
template<class T>
const SmartPtr<T>&
ptr_cast(const AnyPtr& a){
	if(can_cast<T>(a)){
		return unchecked_ptr_cast<T>(a);
	}
	else{
		return nul<T>();
	}
}

//@}

//////////////////////////////////////////

template<class T>
inline typename Caster<T>::type 
tricky_cast(const AnyPtr& a, void (*f)(T)){
	return cast<T>(a);
}

}

#endif // XTAL_CAST_H_INCLUDE_GUARD
