/** \file src/xtal/xtal_environment.h
* \brief src/xtal/xtal_environment.h
*/

#ifndef XTAL_ENVIRONMENT_H_INCLUDE_GUARD
#define XTAL_ENVIRONMENT_H_INCLUDE_GUARD

#pragma once

namespace xtal{

/// \name Xtal̏Aj
//@{

/**
* \brief AP[^Cu
*/
class AllocatorLib{
public:
	virtual ~AllocatorLib(){}
	virtual void* malloc(std::size_t size){ return std::malloc(size); }
	virtual void free(void* p, std::size_t size){ std::free(p); }

	virtual void* malloc_align(std::size_t size, std::size_t alignment);
	virtual void free_align(void* p, std::size_t size, std::size_t alignment);

	virtual void out_of_memory(){}
};

/**
* \brief R[hCu
*/
class ChCodeLib{
public:
	virtual ~ChCodeLib(){}

	virtual int_t ch_len(char_t lead){ return 1; }
	virtual int_t ch_len2(const char_t* str){ return ch_len(*str); }
	virtual int_t ch_inc(const char_t* data, int_t data_size, char_t* dest, int_t dest_size);
	virtual int_t ch_cmp(const char_t* a, int_t asize, const char_t* b, int_t bsize);
};

/**
* \brief XbhCu
*/
class ThreadLib{
public:
	virtual ~ThreadLib(){}

	virtual void* new_thread(){ return 0; }
	virtual void delete_thread(void* thread_object){}
	virtual void start_thread(void* thread_object, void (*callback)(void*), void* data){ callback(data); }
	virtual void join_thread(void* thread_object){}

	virtual void* new_mutex(){ return 0; }
	virtual void delete_mutex(void* mutex_object){}
	virtual void lock_mutex(void* mutex_object){}
	virtual void unlock_mutex(void* mutex_object){}

	virtual void yield(){}
	virtual void sleep(float_t sec){}
};

/**
* \brief Wo̓Cu
*/
class StdStreamLib{
public:
	virtual ~StdStreamLib(){}

	virtual void* new_stdin_stream(){ return 0; }
	virtual void delete_stdin_stream(void* stdin_stream_object){}
	virtual uint_t read_stdin_stream(void* stdin_stream_object, void* dest, uint_t size){ return 0; }

	virtual void* new_stdout_stream(){ return 0; }
	virtual void delete_stdout_stream(void* stdout_stream_object){}
	virtual uint_t write_stdout_stream(void* stdout_stream_object, const void* src, uint_t size){ return 0; }

	virtual void* new_stderr_stream(){ return 0; }
	virtual void delete_stderr_stream(void* stderr_stream_object){}
	virtual uint_t write_stderr_stream(void* stderr_stream_object, const void* src, uint_t size){ return 0; }
};

/**
* \brief t@CVXeCu
*/
class FilesystemLib{
public:
	virtual ~FilesystemLib(){}

	virtual bool is_directory(const char_t* path){ return false; }

	virtual void* new_file_stream(const char_t* path, const char_t* flags){ return 0; }
	virtual void delete_file_stream(void* file_stream_object){}
	virtual uint_t read_file_stream(void* file_stream_object, void* dest, uint_t size){ return 0; }
	virtual uint_t write_file_stream(void* file_stream_object, const void* src, uint_t size){ return 0; }
	virtual void seek_file_stream(void* file_stream_object, uint_t pos){}
	virtual uint_t tell_file_stream(void* file_stream_object){ return 0; }
	virtual bool end_file_stream(void* file_stream_object){ return true; }
	virtual uint_t size_file_stream(void* file_stream_object){ return 0; }
	virtual void flush_file_stream(void* file_stream_object){}

	virtual void* new_entries(const char_t* path){ return 0; }
	virtual void delete_entries(void* entries_object){}
	virtual const char_t* next_entries(void* entries_object){ return 0; }
	virtual void break_entries(void* entries_object){}
};

/**
* \brief gpCu̎ŵ߂̍\
*/
struct Setting{
	AllocatorLib* allocator_lib;
	ChCodeLib* ch_code_lib;
	ThreadLib* thread_lib;
	StdStreamLib* std_stream_lib;
	FilesystemLib* filesystem_lib;

	/**
	* \brief قƂǉȂݒ肷B
	*/
	Setting();
};

/**
* \brief Xtals쐬AAJgɐݒ肷B
*/
void initialize(const Setting& setting);

/**
* \brief JgXtalsjB
*/
void uninitialize();

/**
* \brief ݂̃JgXtals擾B
*/
Environment* environment();

/**
* \brief JgXtalsݒ肷B
*/
void set_environment(Environment* e);

//@}

ThreadLib* thread_lib();

StdStreamLib* std_stream_lib();

FilesystemLib* filesystem_lib();

/////////////////////////////////////////////////////

/// \name 
//@{

/**
* \brief [U[o^AP[g֐gămۂB
*/
void* xmalloc(size_t size);

/**
* \brief [U[o^fAP[g֐găB
*/
void xfree(void* p, size_t size);

/**
* \brief mۂXR[vɕ߂邽߂̃[eBeBNX
*/
struct XMallocGuard{
	/**
	* \brief mۂɍ\zRXgN^
	*/
	XMallocGuard():p(0){}
	
	/**
	* \brief sizẽmۂRXgN^
	*/
	explicit XMallocGuard(uint_t size):p(xmalloc(size)), sz(size){}

	/**
	* \brief 
	*/
	~XMallocGuard(){ if(p)xfree(p, sz); }
public:	

	/**
	* \brief mۂ
	* ܂łɊmۂĂ͉
	*/
	void malloc(size_t size){ if(p)xfree(p, sz); p = xmalloc(size); sz = size; }

	/**
	* \brief ̐擪AhXԂ
	*/
	void* get(){ return p; }

	/**
	* \brief ̐擪AhXԂ
	*/
	void* release(){
		void* temp = p;
		p = 0;
		return temp; 
	}

	/**
	* \brief mۂĂ郁̃TCYԂ
	*/
	uint_t size(){ return sz; }

private:
	void* p;
	uint_t sz;

	XTAL_DISALLOW_COPY_AND_ASSIGN(XMallocGuard);
};

#define XTAL_PROTECT if(const ::xtal::Protect& xtalprotect = setjmp(::xtal::protect().buf)!=0)

#define XTAL_OUT_OF_MEMORY else 

//@}

/**
* \internal
* \brief jmp_buf߂\
*/
struct JmpBuf{
	jmp_buf buf;
};

JmpBuf& protect();

void reset_protect();

bool ignore_memory_assert();

struct Protect{
	bool pass;
	Protect(bool pass):pass(pass){}
	~Protect(){ reset_protect(); }
	operator bool() const{ return !pass; }
};

/////////////////////////////////////////////////////

/**
* \xbind lib::builtin
* \brief ɏIK[xWRNVs
*
* gc֐菈ȂK[xWRNg֐
* Q[Ŗt[܂߂ɌĂԂƂȂǂɎgB
*/
void lw_gc();

/**
* \xbind lib::builtin
* \brief K[xWRNVs
*
* قǎԂ͂ȂASɃS~łȂK[xWRNg֐
* ႦΏzQƂ͂ł͌młȂB
*/
void gc();

/**
* \xbind lib::builtin
* \brief zIuWFNgtK[xWRNVs
*
* Ԃ͂邪A݃S~ƂȂĂ̂͂ȂׂSĉK[xWRNg֐
* zQƂmłB
*/
void full_gc();

/**
* \xbind lib::builtin
* \brief K[xWRNV𖳌
*
* gcfull_gčĂяo𖳌֐B
* łꂪĂяoꂽLĂAĂяoenable_gcĂтȂƃK[xWRNV͗LɂȂȂ
*/
void disable_gc();

/**
* \xbind lib::builtin
* \brief K[xWRNVL

* disable_gcĂ΂ꂽ񐔂ƓĂяoƃK[xWRNVLɂȂ
*/
void enable_gc();

void set_gc_stress(bool b);

uint_t alive_object_count();

const AnyPtr& alive_object(uint_t i);

/////////////////////////////////////////////////////

/**
* \internal
* \brief keyɑΉC++̃NX̃NXIuWFNgԂB
*/
const ClassPtr& cpp_class(CppClassSymbolData* key);
const ClassPtr& cpp_class(int_t index);

/**
* \brief NXTɑΉC++̃NX̃NXIuWFNgԂB
*/
template<class T>
inline const ClassPtr& cpp_class(){
	return cpp_class(&CppClassSymbol<T>::value);
}

/////////////////////////////////////////////////////

/**
* \internal
* \brief T^̃IuWFNgo
*/
const AnyPtr& cpp_value(CppValueSymbolData* key);

/**
* \brief T^̃IuWFNgo
*/
template<class T>
const SmartPtr<T>& cpp_value(){
	return unchecked_ptr_cast<T>(cpp_value(&CppValueSymbol<T>::value));
}

/////////////////////////////////////////////////////

/**
* \internal
* \brief NX̃ooB
*/
const AnyPtr& cache_member(Base* target_class, const IDPtr& primary_key, const AnyPtr& secondary_key, int_t& accessibility);

/**
* \internal
* \brief NX̌p֌W𒲂ׂB
*/
bool cache_is(const AnyPtr& target_class, const AnyPtr& klass);

/**
* \internal
* \brief NX̃RXgN^LbVĂ邩璲ׂB
*/
bool cache_ctor(const AnyPtr& target_class, int_t kind);

/**
* \internal
* \brief LbVB
*/
void clear_cache();

/**
* \internal
* \brief o[̃LbVe[uɓo^Ăf[^𖳌ɂB
*/
void invalidate_cache_member();

/**
* \internal
* \brief p֌W̃LbVe[uɓo^Ăf[^𖳌ɂB
*/
void invalidate_cache_is();

/**
* \internal
* \brief NX̃RXgN^̃LbVe[uɓo^Ăf[^𖳌ɂB
*/
void invalidate_cache_ctor();

/////////////////////////////////////////////////////

/**
* \internal
* \brief VMachineCX^X^B
*/
VMachinePtr vmachine_take_over();

/**
* \internal
* \brief VMachineCX^XԋpB
*/
void vmachine_take_back(const VMachinePtr& vm);

void vmachine_swap_temp();

/////////////////////////////////////////////////////

/// \name CuIuWFNg擾֐
// @{

/**
* \brief builtinIuWFNgԂ
*/
const ClassPtr& builtin();

/**
* \brief libIuWFNgԂ
*/
const LibPtr& lib();

/**
* \brief globalIuWFNgԂ
*/
const ClassPtr& global();

/**
* \brief stdinXg[IuWFNgԂ
*/
const StreamPtr& stdin_stream();

/**
* \brief stdoutXg[IuWFNgԂ
*/
const StreamPtr& stdout_stream();

/**
* \brief stderrXg[IuWFNgԂ
*/
const StreamPtr& stderr_stream();

/**
* \brief VMachinePtrIuWFNgԂ
*
* O[oVMachinePtrIuWFNgԂB
* Xbhɂ̃O[oVMachinePtrIuWFNg݂͑B
*/
const VMachinePtr& vmachine();

const VMachinePtr& vmachine_checked();

const VMachinePtr& setup_call();
const VMachinePtr& setup_call(int_t need_result_count);

// @}

/**
* \brief C^[ςݕɕϊ
*/
IDPtr intern(const char_t* str);

/**
* \brief NULI[CC^[ςݕɕϊ
*
* \param str NULLI[
*/
IDPtr intern(const char8_t* str);

/**
* \brief C񂩂size̒oC^[ςݕɕϊ
*
*/
IDPtr intern(const char_t* str, uint_t size);

/**
* \brief C񃊃eC^[ςݕɕϊ
*
*/
IDPtr intern(const StringLiteral& str);

/**
* \brief beginlast܂ł̕ŃC^[ςݕɕϊ
*
* [begin, last)
*/
IDPtr intern(const char_t* begin, const char_t* last);

/**
* \brief StringC^[ςݕɕϊ
*
*/
IDPtr intern(const StringPtr& name);		


struct StringConv{
	XMallocGuard memory;
	StringConv(const char8_t* str);
};


/**
* \brief bN
*/
void xlock();

/**
* \brief AbN
*/
void xunlock();

struct XUnlock{
	XUnlock(int){ xunlock(); }
	~XUnlock(){ xlock(); }
	operator bool() const{ return true; }
private:
	XUnlock(const XUnlock&);
	void operator =(const XUnlock&);
};

void register_thread(Environment*);

void unregister_thread(Environment*);

/**
* \brief eLXg}bvԂ
*/
const MapPtr& text_map();

/**
* \internal
* \brief 擪oCgāÃ}`oCg𒲂ׂB
*
* }CiX̒lԂꂽꍇAŒᕶԂB
* -2̏ꍇAŒ2ȏ̕ŁA{̒2ڂǂ܂ȂƔfłȂAƂӖƂȂB
*/
int_t ch_len(char_t lead);

/**
* \internal
* \brief }`oCg𒲂ׂB
*
* int_t ch_len(char_t lead)Ă΂ꂽA}CiX̒lԂꍇɑČĂԁB
* ch_len-2̒lԂ́Astr̐ɂ͍Œ2oCg̃f[^i[邱ƁB
*/
int_t ch_len2(const char_t* str);

/**
* \internal
* \brief ̕Ԃ
*
* Ⴆ a nꍇAb Ԃ
*/
int_t ch_inc(const char_t* data, int_t data_size, char_t* dest, int_t dest_size);

/**
* \internal
* \brief ̑召
*
* ̒l a ̕̕b̕菬
* 0̒l 
* ̒l b̕̕a̕菬
*/
int_t ch_cmp(const char_t* a, uint_t asize, const char_t* b, uint_t bsize);

/**
* \internal
* \brief Zq̖O킷ID̔zԂ
*/
const IDPtr* id_op_list();

StreamPtr open(const StringPtr& file_name, const StringPtr& mode);

#ifndef XTAL_NO_PARSER

/// \name RpCn֐
//@{

/**
* \xbind lib::builtin
* \brief file_namet@CRpCB
* ̖߂lserializeƁAoCgR[h`ŕۑB
* \param file_name XtalXNvgLqꂽt@C̖O
* \return słCodeIuWFNg
*/
CodePtr compile_file(const StringPtr& file_name);

/**
* \xbind lib::builtin
* \brief sourceRpCB
* ̖߂lserializeƁAoCgR[h`ŕۑB
* \param source XtalXNvgLqꂽ
* \return słCodeIuWFNg
*/
CodePtr compile(const AnyPtr& source);

/**
* \xbind lib::builtin
* \brief sourceevalpɃRpCB
* \param source XtalXNvgLqꂽ
* \return VMachine::evalɓnƂ̂łCodeIuWFNg
*/
CodePtr eval_compile(const AnyPtr& source);

/**
* \xbind lib::builtin
* \brief file_namet@CRpCĎsB
* \param file_name XtalXNvgLqꂽt@C̖O
* \return XNvgreturnꂽl
*/
AnyPtr load(const StringPtr& file_name);

void set_require_source_hook(const AnyPtr& hook);

CodePtr require_source(const StringPtr& name);

AnyPtr require(const StringPtr& name);

//@}

CodePtr source(const char_t* src, int_t size);

void exec_source(const char_t* src, int_t size);

#endif

CodePtr compiled_source(const void* src, int_t size);

void exec_compiled_source(const void* src, int_t size);

}

#endif // XTAL_ENVIRONMENT_H_INCLUDE_GUARD
